package com.yihu.quota.service.view;

import com.fasterxml.jackson.core.type.TypeReference;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.yihu.ehr.elasticsearch.ElasticSearchPool;
import com.yihu.ehr.query.BaseJpaService;
import com.yihu.quota.dao.view.ViewDao;
import com.yihu.quota.dao.view.ViewDimensionDao;
import com.yihu.quota.dao.view.ViewQuotaDao;
import com.yihu.quota.dao.view.ViewQuotaFilterDao;
import com.yihu.quota.model.cube.Cube;
import com.yihu.quota.model.view.View;
import com.yihu.quota.model.view.ViewDimension;
import com.yihu.quota.model.view.ViewQuota;
import com.yihu.quota.model.view.ViewQuotaFilter;
import com.yihu.quota.vo.ViewQuotaFilterModel;
import org.apache.commons.lang3.StringUtils;
import org.apache.log4j.Logger;
import org.elasticsearch.action.search.SearchRequestBuilder;
import org.elasticsearch.action.search.SearchResponse;
import org.elasticsearch.client.transport.TransportClient;
import org.elasticsearch.index.query.BoolQueryBuilder;
import org.elasticsearch.index.query.QueryBuilder;
import org.elasticsearch.index.query.QueryBuilders;
import org.elasticsearch.search.aggregations.AggregationBuilders;
import org.elasticsearch.search.aggregations.Aggregations;
import org.elasticsearch.search.aggregations.metrics.cardinality.Cardinality;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;


/**
 * @author janseny
 * @date 2018年9月26日
 */
@Service
public class ViewService extends BaseJpaService<View, ViewDao> {

    private static final Logger logger = Logger.getLogger(ViewService.class);

    @Autowired
    private ViewDao viewDao;
    @Autowired
    private ViewDimensionDao viewDimensionDao;
    @Autowired
    private ViewQuotaDao viewQuotaDao;
    @Autowired
    private ViewQuotaFilterDao viewQuotaFilterDao;
    @Autowired
    private ElasticSearchPool elasticSearchPool;
    @Autowired
    private ObjectMapper objectMapper;

    public View findOne(String id) {
        return viewDao.findOne(id);
    }

    /**
     * 获取视图完整规则
     *
     * @param viewCode 视图编码
     * @return 视图规则
     */
    public View getViewRule(String viewCode) {
        View view = viewDao.findByCode(viewCode);
        // 视图默认过滤条件
        view.setViewFilterList(viewQuotaFilterDao.findByRelationId(view.getId()));
        // 视图各组组内顶层行维度（升序）
        List<ViewDimension> groupTopRowDimensionList = viewDimensionDao.getGroupTopRowDimensionList(view.getId());
        view.setGroupTopRowDimensionList(groupTopRowDimensionList);
        // 视图各组组内其他行维度（升序）
        Map<Integer, List<ViewDimension>> groupOtherRowDimensionsMap = new HashMap<>(16);
        for (ViewDimension viewDimension : groupTopRowDimensionList) {
            List<ViewDimension> otherRowDimensionList = viewDimensionDao.getGroupOtherRowDimensionList(view.getId(), viewDimension.getGroupRow());
            groupOtherRowDimensionsMap.put(viewDimension.getGroupRow(), otherRowDimensionList);
        }
        view.setGroupOtherRowDimensionsMap(groupOtherRowDimensionsMap);
        // 视图列维度（升序）
        view.setColDimensionList(viewDimensionDao.getColDimensionList(view.getId()));
        // 视图指标
        view.setViewQuotaList(viewQuotaDao.findByViewId(view.getId()));
        for (ViewQuota viewQuota : view.getViewQuotaList()) {
            List<ViewQuotaFilter> quotaFilterList = viewQuotaFilterDao.findByRelationId(viewQuota.getId());
            viewQuota.setQuotaFilterList(quotaFilterList);
        }
        return view;
    }

    /**
     * 获取视图统计结果
     *
     * @param viewCode
     * @param filterModelList
     */
    public Aggregations statViewResult(String viewCode, List<ViewQuotaFilterModel> filterModelList) throws IOException {
        // 获取视图规则
        View view = viewDao.findByCode(viewCode);
        String[] esIndices = view.getEsIndex().split(",");
        String[] esTypes = view.getEsType().split(",");

        TransportClient transportClient = elasticSearchPool.getClient();
        SearchRequestBuilder searchRequestBuilder = transportClient.prepareSearch(esIndices).setTypes(esTypes).setSize(0);

        // 组装页面传入的过滤条件
        BoolQueryBuilder viewBoolQueryBuilder = QueryBuilders.boolQuery();
        joinViewFilters(viewBoolQueryBuilder, filterModelList);
        // 组装视图默认过滤条件
        String viewFilterStr = objectMapper.writeValueAsString(view.getViewFilterList());
        List<ViewQuotaFilterModel> viewFilterModelList = objectMapper.readValue(viewFilterStr, new TypeReference<List<ViewQuotaFilterModel>>() {
        });
        joinViewFilters(viewBoolQueryBuilder, viewFilterModelList);
        searchRequestBuilder.setQuery(viewBoolQueryBuilder);

        // 组装聚合
        List<ViewDimension> groupTopRowDimensionList = view.getGroupTopRowDimensionList();
        if (groupTopRowDimensionList.size() != 0) {
            int indexCount = esIndices.length;
            for (ViewDimension viewDimension : groupTopRowDimensionList) {
                // 有多个多维度数据集时，设定顶层聚合的数据来自哪个多维数据集
                if (indexCount > 1) {
                    AggregationBuilders.filter(viewDimension.getDimensionCode());
                }
            }
        }

        SearchResponse searchResponse = searchRequestBuilder.get();
        Aggregations aggregations = searchResponse.getAggregations();
        transportClient.close();

        return aggregations;
    }

    /**
     * 拼接视图的过滤条件
     *
     * @param boolQueryBuilder
     * @param viewFilterModelList
     */
    private void joinViewFilters(BoolQueryBuilder boolQueryBuilder, List<ViewQuotaFilterModel> viewFilterModelList) {
        if (viewFilterModelList.size() != 0) {
            for (ViewQuotaFilterModel filterModel : viewFilterModelList) {
                String field = filterModel.getDimensionCode();
                String value = filterModel.getFilterValue();
                // 属于与否，TODO 需要判断字段类型，转换 value 的类型。
                if ("belong".equals(filterModel.getCompareType())) {
                    QueryBuilder queryBuilder = QueryBuilders.termQuery(field, value);
                    joinBoolQuery(boolQueryBuilder, queryBuilder, filterModel.getRelationType());
                } else if ("not_belong".equals(filterModel.getCompareType())) {
                    QueryBuilder queryBuilder = QueryBuilders.termQuery(field, value);
                    joinBoolQuery(boolQueryBuilder, queryBuilder, "not");
                }
                // 为空与否
                else if ("null".equals(filterModel.getCompareType())) {
                    QueryBuilder queryBuilder = QueryBuilders.missingQuery(field).nullValue(true);
                    joinBoolQuery(boolQueryBuilder, queryBuilder, filterModel.getRelationType());
                } else if ("not_null".equals(filterModel.getCompareType())) {
                    QueryBuilder queryBuilder = QueryBuilders.missingQuery(field).nullValue(true);
                    joinBoolQuery(boolQueryBuilder, queryBuilder, "not");
                }
                // 包含与否
                else if ("null".equals(filterModel.getCompareType())) {
                    QueryBuilder queryBuilder = QueryBuilders.matchQuery(field, value);
                    joinBoolQuery(boolQueryBuilder, queryBuilder, filterModel.getRelationType());
                } else if ("not_null".equals(filterModel.getCompareType())) {
                    QueryBuilder queryBuilder = QueryBuilders.matchQuery(field, value);
                    joinBoolQuery(boolQueryBuilder, queryBuilder, "not");
                }
                // 范围，TODO 需要判断字段类型，转换 value 的类型。
                else if ("gt".equals(filterModel.getCompareType())) {
                    QueryBuilder queryBuilder = QueryBuilders.rangeQuery(field).gt(value);
                    joinBoolQuery(boolQueryBuilder, queryBuilder, filterModel.getRelationType());
                } else if ("gte".equals(filterModel.getCompareType())) {
                    QueryBuilder queryBuilder = QueryBuilders.rangeQuery(field).gte(value);
                    joinBoolQuery(boolQueryBuilder, queryBuilder, filterModel.getRelationType());
                } else if ("lt".equals(filterModel.getCompareType())) {
                    QueryBuilder queryBuilder = QueryBuilders.rangeQuery(field).lt(value);
                    joinBoolQuery(boolQueryBuilder, queryBuilder, filterModel.getRelationType());
                } else if ("lte".equals(filterModel.getCompareType())) {
                    QueryBuilder queryBuilder = QueryBuilders.rangeQuery(field).lte(value);
                    joinBoolQuery(boolQueryBuilder, queryBuilder, filterModel.getRelationType());
                }
            }
        }
    }

    private void joinBoolQuery(BoolQueryBuilder boolQueryBuilder, QueryBuilder queryBuilder, String relationType) {
        if ("and".equals(relationType)) {
            boolQueryBuilder.must(queryBuilder);
        } else if ("or".equals(relationType)) {
            boolQueryBuilder.should(queryBuilder);
        } else if ("not".equals(relationType)) {
            boolQueryBuilder.mustNot(queryBuilder);
        }
    }

}